import os
import json
import asyncio
import websockets
import urllib

from typing import Union, Set, Optional
from pydantic import AnyUrl

from datapools.producer import BaseProducer
from datapools.scheduler import CrawlerScheduler
from datapools.worker import CrawlerWorker
from datapools.common.logger import logger

from datapools.common.types import (
    BaseProducerSettings,
    SchedulerSettings,
    WorkerSettings,
)


"""
In cli mode `plugins` keys are used to define WorkerSettings.USE_ONLY_PLUGINS and ADDITIONAL_PLUGINS fields:
So only listed plugins will be loaded.

`queue.connection_url` expects Rabbitmq running locally. 
Run it as `docker run --rm -p 15672:15672 -p 5672:5672 rabbitmq:3.10.7-management` for example
"""
default_config = {
    "backend_api_url": "https://openlicense.ai/openlicense/api/",
    "queue":{
      "connection_url": "amqp://guest:guest@localhost:5672",
      "size_limit": 1
    },
    "storage":{
      "worker":{
        "path": "/tmp/datapools_worker_storage"
      },
      "producer":{
        "path": "/tmp/datapools_producer_storage"
      }
    },
    "plugins":{
      "s3":{
        "aws_access_key_id": None,
        "aws_secret_access_key": None
      },
      "google_drive":{
        "api_key": ""
      },
      "imageshack":{
      }
    }
}

#async def crawl( hint_urls: Union[AnyUrl, Set[AnyUrl]], config_file: Optional[str] = None):
async def crawl( datapool_id: int, config_file: Optional[str] = None):
    
    #loading config from json or using default one

    if config_file is not None:
        with open(config_file,'r') as f:
            config = json.load(f)
    else:
        config = default_config
        
    # initializing and starting Scheduler
    s_cfg = SchedulerSettings()
    s_cfg.load( config )  
    s_cfg.CLI_MODE = True  
    #cfg.CLI_HINT_URLS = hint_urls if type(hint_urls) is set else { hint_urls }
    scheduler = CrawlerScheduler(s_cfg)
    scheduler.run()
        
    # initializing and starting Worker
    w_cfg = WorkerSettings()
    w_cfg.load( config )
    w_cfg.CLI_MODE = True    
    if not os.path.exists(w_cfg.STORAGE_PATH):
        os.mkdir(w_cfg.STORAGE_PATH)

    plugins_list = [ plugin_name for plugin_name in config[ 'plugins' ] ]
    w_cfg.USE_ONLY_PLUGINS = plugins_list
    w_cfg.ADDITIONAL_PLUGINS = plugins_list

    worker = CrawlerWorker(w_cfg)
    worker.run()

    # initializing and starting Producer
    p_cfg = BaseProducerSettings()
    p_cfg.load( config )
    p_cfg.CLI_MODE = True
    if not os.path.exists(p_cfg.STORAGE_PATH):
        os.mkdir(p_cfg.STORAGE_PATH)    
    
    producer = BaseProducer(p_cfg)
    producer.run()
    
    # get datapool contents stream from the backend
    o = urllib.parse.urlparse(f"{config[ 'backend_api_url' ]}get-datapool-stream")
    if o.scheme == 'http':
        url = o._replace(scheme='ws').geturl()
    elif o.scheme == 'https':
        url = o._replace(scheme='wss').geturl()
    logger.info( f'{url=}')

    async with websockets.connect(url) as ws:
        await ws.send(json.dumps({"id": datapool_id}))
        try:
            while True:
                msg = json.loads(await ws.recv())
                logger.info( f'{msg=}')
                await scheduler.add_hint( msg['url'] )
                #break
        except websockets.exceptions.ConnectionClosed:
            logger.info( 'connection closed')
            
    logger.info( 'pushing end marker')
    await scheduler.add_hint( '' )  #telling scheduler that this is the end of stream
        
    await asyncio.gather(scheduler.wait(), worker.wait(), producer.wait())
    logger.info( 'waited all')
    
    await asyncio.gather(scheduler.stop(), worker.stop(), producer.stop())
    logger.info( 'stopped all')