from enum import Enum
import json
from typing import Any, List, Optional, Set, Union

from pydantic import BaseModel, BaseSettings

# from pydantic_settings import BaseSettings

QUEUE_WORKER_TASKS = "worker_tasks"
QUEUE_REPORTS = "worker_reports"
QUEUE_EVAL_TASKS = "eval_tasks"
QUEUE_TOPICS = "topics"

RABBITMQ_HOST = "rabbitmq.crawler"
RABBITMQ_PORT = "5672"

DEFAULT_CONNECTION_URL: str = "amqp://guest:guest@{host}:{port}/".format(host=RABBITMQ_HOST, port=RABBITMQ_PORT)

class InvalidUsageException(Exception):
    pass

 
class SchedulerSettings(BaseSettings):
    QUEUE_CONNECTION_URL: str = DEFAULT_CONNECTION_URL
    BACKEND_API_URL: str = "https://openlicense.ai/openlicense/api/"
    BACKEND_HINTS_PERIOD: int = 10  # seconds
    # cli settings
    CLI_HINT_URLS: Optional[Set[str]] = None
    CLI_MODE: bool = False

    def load(self, config:dict):
        """fills self with json config"""
        # Queue
        queue = config.get('queue',{})
        queue_connection_url = queue.get('connection_url')
        if queue_connection_url is not None:
            self.QUEUE_CONNECTION_URL = queue_connection_url
            
        api_url = config.get('backend_api_url')
        if api_url is not None:
            self.BACKEND_API_URL = api_url


class WorkerSettings(BaseSettings):
    QUEUE_CONNECTION_URL: str = DEFAULT_CONNECTION_URL
    TODO_QUEUE_SIZE: int = 1
    
    ATTEMPTS_PER_URL: int = 3
    ATTEMPTS_DELAY: int = 5  # seconds
    STORAGE_PATH: str = "/storage/"

    # None: access is configured on AWS, bucket is NOT PUBLIC
    # "": bucket is PUBLIC
    # S3_IMAGESHACK_ACCESS_KEY: Optional[str] = None  
    # S3_IMAGESHACK_ACCESS_SECRET: Optional[str] = None

    # GOOGLE_DRIVE_API_KEY: str = ""

    CLI_MODE: bool = False

    USE_ONLY_PLUGINS: Optional[List[str]] = None
    ADDITIONAL_PLUGINS: Optional[List[str]] = None
    
    plugins_config: dict = {}
    def fload(self, json_path):
        """fills self using json file path"""
        with open( json_path, 'r' ) as fp:
            config = json.load(fp)  #expects dict output
            self.load( config )

    def load(self, config:dict):
        """fills self with json config"""
        # Queue
        queue = config.get('queue',{})
        queue_connection_url = queue.get('connection_url')
        if queue_connection_url is not None:
            self.QUEUE_CONNECTION_URL = queue_connection_url
        queue_size_limit = queue.get( 'size_limit')
        if queue_size_limit is not None:
            self.TODO_QUEUE_SIZE = queue_size_limit
        
        # Storage
        storage = config.get( 'storage', {})
        worker_storage = storage.get( 'worker', {} )
        storage_path = worker_storage.get( 'path' )
        if storage_path is not None:
            self.STORAGE_PATH = storage_path
            
        # plugins
        self.plugins_config = config.get('plugins',{})
            


class BaseProducerSettings(BaseSettings):
    QUEUE_CONNECTION_URL: str = DEFAULT_CONNECTION_URL
    BACKEND_API_URL: str = "https://openlicense.ai/openlicense/api/"
    STORAGE_PATH: str = "/storage/"
    WORKER_STORAGE_PATH: str = "/worker_storage/"

    CLI_MODE: bool = False

    def load(self, config:dict):
        """fills self with json config"""
        # Queue
        queue = config.get('queue',{})
        queue_connection_url = queue.get('connection_url')
        if queue_connection_url is not None:
            self.QUEUE_CONNECTION_URL = queue_connection_url
        
        # Storage
        storage = config.get( 'storage', {})
        worker_storage = storage.get( 'worker', {} )
        storage_path = worker_storage.get( 'path' )
        if storage_path is not None:
            self.WORKER_STORAGE_PATH = storage_path
        producer_storage = storage.get( 'producer', {} )
        storage_path = producer_storage.get( 'path' )
        if storage_path is not None:
            self.STORAGE_PATH = storage_path
            

class CrawlerHintURLStatus(Enum):
    Unprocessed = 0
    Success = 1
    Failure = 2
    Processing = 3
    Rejected = 4


class DatapoolContentType(Enum):
    Text = "Text"
    Image = "Image"
    Video = "Video"
    Audio = "Audio"

    def __str__(self):
        return self.value

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        if self.value == DatapoolContentType.Text:
            return 1
        if self.value == DatapoolContentType.Image:
            return 2
        if self.value == DatapoolContentType.Video:
            return 3
        if self.value == DatapoolContentType.Audio:
            return 4
        raise Exception(
            f"Not supported DatapoolContentType __hash__ {self.value}"
        )


class CrawlerContent(BaseModel):
    tag_id: Optional[str]
    copyright_tag_id: Optional[str]
    platform_tag_id: Optional[str]
    type: DatapoolContentType
    storage_id: Any
    url: str

    def to_dict(self):
        res = self.__dict__
        res["type"] = res["type"].value
        return res


class CrawlerBackTask(BaseModel):
    url: str

    def to_dict(self):
        res = self.__dict__
        return res


class CrawlerNop:
    pass


class Evaluation(BaseModel):
    nsfw: bool
    score: float
    weight: float
    embeddings: Optional[List[float]]

    # def to_dict(self):
    #     return {
    #         'nsfw': self.nsfw,
    #         'score': self.score,
    #         'weight': self.weight
    #     }


